# -*- coding: utf-8 -*-
"""Contains database file class and container classes derived from it."""

import re
import warnings
from pathlib import Path

from .exceptions import NotInLibraryError, ConstructorError
from .database import spextra_database
from .libraries import Library, SpecLibrary, FilterSystem, ExtCurvesLibrary


__all__ = ["SpectrumContainer", "ExtCurveContainer", "FilterContainer"]

NAME_REGEX = re.compile(r"^(?P<libname>\w+(/\w+)*)/(?P<basename>[\w\.\-]+)$")


class DBFile:
    """Base class for database files."""

    library = None  # necessary to avoid access before init ran (REALLY??)
    _subclass_str = "DBFile"
    _subclass_library = Library

    def __init__(self, name: str):
        # To avoid name conflict with backwards-compatible meaning
        # In the future, just call it .name
        self._name = name
        name_match = NAME_REGEX.match(name)
        if not name_match:
            raise ConstructorError(
                f"{self._subclass_str} name '{name}' must match the pattern "
                "'library_name/file_name'.")

        self.basename = name_match["basename"]
        self.library = self._subclass_library(name_match["libname"])
        # TODO: maybe an info log if not in DB?

    @property
    def is_in_library(self) -> bool:
        """Return True if file is part of the parent library."""
        return self.basename in self.library

    @property
    def datafile(self):
        """Name and extension of the file."""
        warnings.warn("The 'name plus file extension' property will be moved "
                      "to .filename in the next minor pre-release, making "
                      ".datafile deprecated.",
                      PendingDeprecationWarning, 2)
        return self.basename + self.library.file_extension

    @property
    def path(self) -> Path:
        """Path to the cached file."""
        if not self.is_in_library:
            raise NotInLibraryError(f"{self._subclass_str} '{self.basename}' "
                                    "not in library")
        return spextra_database.fetch(f"{self.library.path}/{self.datafile}")

    @property
    def filename(self):
        """Deprecated feature."""
        warnings.warn("The .filename property is deprecated and will refer to "
                      "the full name incl. file extension in future versions. "
                      "For the absolute file path, please use .path instead!",
                      DeprecationWarning, 2)
        return self.path

    @property
    def name(self):
        """Deprecated feature."""
        warnings.warn("Accessing the library name directly via the .name "
                      "property is ambiguous. In future versions, .name will "
                      "refer to the .basename attribute + library_name. "
                      "Please use the more explicit .library.name instead!",
                      DeprecationWarning, 2)
        return self.library.name

    @property
    def description(self) -> str:
        """Description or comment of the file as defined in the library."""
        return self.library[self.basename]

    def remove(self) -> None:
        """Remove the file."""
        self.path.unlink()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._name!r})"

    def __str__(self) -> str:
        return f"{self._subclass_str} '{self._name}' ({self.description})"

    def __getattr__(self, key: str):
        """Allow (hacky) direct access to attributes of library."""
        if not key.startswith("__") and hasattr(self.library, key):
            warnings.warn(f"Accessing library attributes like '{key}' "
                          "directly is deprecated and will raise an error in "
                          "future versions. Please use the more explicit "
                          f".library.{key} instead!", DeprecationWarning, 2)
        return getattr(self.library, key)


class SpectrumContainer(DBFile):
    """Container for a spectral template file."""

    _subclass_str = "Spectral template"
    _subclass_library = SpecLibrary

    @property
    def template(self) -> str:
        """Name of the spectrum template."""
        return self._name

    @property
    def template_name(self):
        """Deprecated feature."""
        warnings.warn("The .template_name property is deprecated and will be "
                      "removed in v1.0. Please use the identical .basename "
                      "instead!", DeprecationWarning, 2)
        return self.basename

    @property
    def template_comment(self):
        """Deprecated feature."""
        warnings.warn("The .template_comment property is deprecated and will be "
                      "removed in v1.0. Please use the identical .description "
                      "instead!", DeprecationWarning, 2)
        return self.description


class FilterContainer(DBFile):
    """Container for a filter curve file."""

    _subclass_str = "Filter"
    _subclass_library = FilterSystem

    @property
    def filter_name(self) -> str:
        """Name of the filter."""
        return self._name

    @property
    def filter_comment(self):
        """Deprecated feature."""
        warnings.warn("The .filter_comment property is deprecated and will be "
                      "removed in v1.0. Please use the identical .description "
                      "instead!", DeprecationWarning, 2)
        return self.description


class ExtCurveContainer(DBFile):
    """Container for extinction curve file."""

    _subclass_str = "Extinction Curve"
    _subclass_library = ExtCurvesLibrary

    @property
    def curve_name(self) -> str:
        """Name of the extinction curve."""
        return self._name

    @property
    def curve_comment(self):
        """Deprecated feature."""
        warnings.warn("The .curve_comment property is deprecated and will be "
                      "removed in v1.0. Please use the identical .description "
                      "instead!", DeprecationWarning, 2)
        return self.description



# -*- coding: utf-8 -*-
"""Spextra-specfic exceptions and errors, for encapsulation."""


__all__ = [
    "SpextraError",
    "NotInDatabaseError",
    "NotInLibraryError",
    "ArgumentError",
    "ConstructorError",
    ]


class SpextraError(Exception):
    """Base class for errors in this Package."""


class NotInDatabaseError(SpextraError, ValueError):
    """Requested library is not in the database."""


class NotInLibraryError(SpextraError, ValueError):
    """Requested file is not in the library."""


class ArgumentError(SpextraError, ValueError):
    """Error in passed arguments."""


class ConstructorError(ArgumentError):
    """Error in constructor arguments."""


class UnitError(SpextraError):
    """Unit not understood by SYNPHOT or astropy."""



# -*- coding: utf-8 -*-
"""Tests for spextra database."""

import urllib
import pytest
import yaml

from spextra.database import Database, DefaultData
from spextra.libraries import SpecLibrary
from spextra.configuration import config

from . import PATH_HERE

path = PATH_HERE / "mocks/index.yml"
with path.open(encoding="utf-8") as file:
    datacont = yaml.safe_load(file)


@pytest.fixture(scope="class")
def mock_database():
    return datacont


@pytest.fixture(scope="class")
def mock_defaultdata():
    return DefaultData.from_yamls()


@pytest.fixture(scope="module")
def url_result():
    url = config.database_url
    return urllib.parse.urlparse(url)


@pytest.mark.parametrize("attribute", ["scheme", "netloc", "path"])
def test_database_url(url_result, attribute):
    """test if a valid url"""
    assert getattr(url_result, attribute)


class TestDatabase:
    @pytest.mark.parametrize("key", ["libraries",
                                     "extinction_curves",
                                     "filter_systems"])
    def test_keys(self, key, mock_database):
        assert key in mock_database
        assert len(mock_database[key])


class TestDefaultData:
    @pytest.mark.parametrize("attr", ["filters", "extcurves", "spectra"])
    def test_attrs(self, attr, mock_defaultdata):
        assert hasattr(mock_defaultdata, attr)


class TestSpecLibrary:
    @pytest.mark.parametrize("library_name", datacont["libraries"])
    def test_name(self, library_name):
        lib = SpecLibrary(library_name)
        assert lib.name == library_name

    def test_templates(self):
        name = "kc96"
        lib = SpecLibrary(name)
        assert "bulge" in lib.keys()

    @pytest.mark.parametrize("library_name", datacont["libraries"])
    @pytest.mark.parametrize("attribute", ["file_extension", "data_type"])
    def test_attr(self, library_name, attribute):
        lib = SpecLibrary(library_name)
        assert hasattr(lib, attribute)



# -*- coding: utf-8 -*-
"""Database for speXtra."""

import shutil
from io import StringIO
from pathlib import Path
from dataclasses import dataclass

import yaml

from astar_utils import NestedMapping

from .downloads import retriever
from .configuration import config

__all__ = ["load_yamldict", "spextra_database", "DEFAULT_DATA"]


class Database(NestedMapping):
    """Contains the database."""

    def __init__(self):
        self.database_url = config.database_url
        self.data_dir = config.cache_dir

        # TODO: change this to yaml constructor once in astar_utils
        super().__init__(load_yamldict("index.yml"))

    def fetch(self, filename: str) -> Path:
        """
        Return absolute path to file or directory, ensuring that it exists.

        If it doesn't exist it will download it from the remote host.
        Otherwise, just look for `relpath`.
        """
        filename = str(filename)
        abspath = self.data_dir / filename

        # don't download folder (.is_dir doesn't work here)
        if not abspath.suffix:
            return abspath

        if abspath.exists():
            return abspath

        abspath = Path(retriever.fetch(filename, progressbar=True))
        return abspath

    def clear_cache(self) -> None:
        """Remove local copies of all files in the database."""
        try:
            shutil.rmtree(self.data_dir)
            print(f"database at {self.data_dir} removed")
        except FileNotFoundError:
            print(f"database at {self.data_dir} doesn't exist")
            raise

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    def __str__(self) -> str:
        with StringIO() as str_stream:
            str_stream.write("Spextra Database:\n"
                             f"  Remote URL: {self.database_url}\n"
                             f"  Local path: {self.data_dir}\n")
            self.write_string(str_stream)
            output = str_stream.getvalue()
        return output


@dataclass
class DefaultData:
    """Small class just to define defaults spectra and filters."""

    filters: dict
    extcurves: dict
    spectra: dict

    @classmethod
    def from_yamls(cls):
        """Construct instance from default yaml files."""
        yamlfiles = ["default_filters.yml",
                     "default_spectra.yml",
                     "default_curves.yml"]
        paths = (
            path_local
            if (path_local := Path(__file__).parent.parent / "database" / fname).exists()
            else Path(retriever.fetch(fname))
            for fname in yamlfiles
        )
        yamls = dict_from_yamls(*paths)

        return cls(yamls["default_filters"],
                   yamls["default_curves"],
                   yamls["default_spectra"])


def load_yamldict(filename: str) -> dict:
    """Fetch YAML file from database and load contents into dict."""
    filepath_local = Path(__file__).parent.parent / "database" / filename
    if filepath_local.exists():
        filepath = filepath_local
    else:
        filepath = retriever.fetch(filename, progressbar=True)
    with open(filepath, encoding="utf-8") as file:
        yamldict = yaml.safe_load(file)
    return yamldict


def dict_from_yamls(*yamls):
    """Construct a nested dict from one or more yaml files."""
    outdict = {}
    for path in yamls:
        with path.open("r", encoding="utf-8") as file:
            outdict[path.stem] = yaml.safe_load(file)
    return outdict


DEFAULT_DATA = DefaultData.from_yamls()
spextra_database = Database()



# -*- coding: utf-8 -*-
"""Contains ldatabase library and subclasses."""

import warnings
import shutil
from pathlib import Path
from collections.abc import Mapping

from synphot.units import validate_unit
from synphot.exceptions import SynphotError

from .exceptions import UnitError, ConstructorError
from .database import spextra_database, load_yamldict


__all__ = ["SpecLibrary", "ExtCurvesLibrary", "FilterSystem"]


class Library(Mapping):
    """
    Class that contains the information of a Library.

    Either spectral of a filter system, extinction curves, etc.

    Base class, not meant to be instantiated directly.
    """

    db_dir = None
    aliases = {
        "items": "items",
        "wave_col": "wave_col",
        "flux_col": "flux_col",
        "flux_unit": "flux_unit",
        }

    def __init__(self, name):
        self.name = name

        if self.is_in_database:
            yamldict = load_yamldict(f"{self.path}/index.yml")
        else:
            yamldict = {}

        self.title = yamldict.get("title", "<untitled>")
        self._items = yamldict.get(self.aliases["items"], {})
        self.spectral_coverage = yamldict.get("spectral_coverage", [])

        self.read_kwargs = self._get_read_kwargs(yamldict)
        self._validate_units()

        self.file_extension = yamldict.get("file_extension", None)

    def _get_read_kwargs(self, yamldict) -> dict:
        read_kwargs = {
            "wave_col": yamldict.get(self.aliases["wave_col"], "WAVELENGTH"),
            "flux_col": yamldict.get(self.aliases["flux_col"], "FLUX"),
            "wave_unit": yamldict.get("wave_unit", "Angstrom"),
            "flux_unit": yamldict.get(self.aliases["flux_unit"], "FLAM"),
        }
        return read_kwargs

    def _validate_units(self) -> None:
        for key in ("wave_unit", "flux_unit"):
            value = self.read_kwargs[key]
            # HACK to understand ext. units
            if value == "Av/E(B-V)":
                value = "mag"
            try:
                self.read_kwargs[key] = validate_unit(value)
            except (SynphotError, ValueError) as err:
                raise UnitError(f"{value} not understood.") from err

    @property
    def is_in_database(self) -> bool:
        """Return True if library is defined in the database instance."""
        return f"!{self.db_dir}.{self.name}" in spextra_database

    @property
    def path(self):
        return f"{self.db_dir}/{self.name}"

    @property
    def url(self):
        pass  # Not yet supported

    @property
    def files(self):
        """Get a list of all filenames in the library."""
        return [f"{fname}{self.file_extension}" for fname in self]

    @property
    def data_type(self) -> str:
        """Determine data type based on known file extensions."""
        if self.file_extension == ".fits":
            return "fits"
        if self.file_extension == ".dat":
            return "ascii"
        return self.file_extension

    def download_all(self) -> None:
        """Download the whole library."""
        # database = Database()
        for key in self:
            spextra_database.fetch(Path(self.name, key))

    def clear_cache(self) -> None:
        """Remove local copies of all files in the library."""
        # database = Database()
        lib_dir = spextra_database.data_dir / self.path
        try:
            shutil.rmtree(lib_dir)
            print(f"library {self.name} removed")
        except FileNotFoundError:
            print(f"library {self.name} doesn't exist")
            raise

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name!r})"

    def __getitem__(self, key):
        return self._items[key]

    def __iter__(self):
        return iter(self._items)

    def __len__(self) -> int:
        return len(self._items)


class SpecLibrary(Library):
    """Library of spectral templates."""

    db_dir = "libraries"
    aliases = {
        "items": "templates",
        "wave_col": "wave_column_name",
        "flux_col": "flux_column_name",
        "flux_unit": "flux_unit",
        }

    def __init__(self, name=None, library_name=None):
        if library_name is not None:
            warnings.warn("Constructing using the 'library_name' argument is "
                          "deprecated and will raise an error in v1.0. Please "
                          "use the more general 'name' argument.",
                          DeprecationWarning, 2)
            name = library_name
        if name is None and library_name is None:
            raise ConstructorError("name must be passed to constructor")

        super().__init__(name)

    @property
    def library_name(self):
        """Deprecated feature."""
        warnings.warn("The alias .library_name is deprecated and will be "
                      "removed in v1.0. Please use .name instead!",
                      DeprecationWarning, 2)
        return self.name

    @property
    def templates(self):
        """Deprecated feature."""
        warnings.warn("The alias .templates is deprecated and will be "
                      "removed in v1.0. Please use .items() instead!",
                      DeprecationWarning, 2)
        return list(self.items())

    @property
    def template_names(self):
        """Deprecated feature."""
        warnings.warn("The alias .template_names is deprecated and will be "
                      "removed in v1.0. Please use the  more general "
                      ".keys() instead!", DeprecationWarning, 2)
        return list(self.keys())

    @property
    def template_comments(self):
        """Deprecated feature."""
        warnings.warn("The properts .template_comments is deprecated and will "
                      "be removed in v1.0.", DeprecationWarning, 2)
        return list(self.values())

    def __str__(self) -> str:
        outstr = (f"Spectral Library '{self.name}': {self.title}\n"
                  f"  spectral coverage: {', '.join(self.spectral_coverage)}\n"
                  f"  wave_unit: {self.read_kwargs['wave_unit']}\n"
                  f"  flux_unit: {self.read_kwargs['flux_unit']}\n"
                  f"  Templates: {', '.join(self.keys())}")
        return outstr


class FilterSystem(Library):
    """Library of filters comprising a filter system."""

    db_dir = "filter_systems"
    aliases = {
        "items": "filters",
        "wave_col": "wave_col",
        "flux_col": "flux_col",
        "flux_unit": "flux_unit",
        }

    def __init__(self, name=None, filter_system=None):
        if filter_system is not None:
            warnings.warn("Constructing using the 'filter_system' argument is "
                          "deprecated and will raise an error in v1.0. Please "
                          "use the more general 'name' argument.",
                          DeprecationWarning, 2)
            name = filter_system
        if name is None and filter_system is None:
            raise ConstructorError("name must be passed to constructor")

        super().__init__(name)

    @property
    def filter_system(self):
        """Deprecated feature."""
        warnings.warn("The alias .filter_system is deprecated and will be "
                      "removed in v1.0. Please use .name instead!",
                      DeprecationWarning, 2)
        return self.name

    @property
    def filters(self):
        """Deprecated feature."""
        warnings.warn("The alias .filters is deprecated and will be "
                      "removed in v1.0. Please use .items() instead!",
                      DeprecationWarning, 2)
        return list(self.items())

    @property
    def filter_names(self):
        """Deprecated feature."""
        warnings.warn("The alias .filter_names is deprecated and will be "
                      "removed in v1.0. Please use the  more general "
                      ".keys() instead!", DeprecationWarning, 2)
        return list(self.keys())

    @property
    def filter_comments(self):
        """Deprecated feature."""
        warnings.warn("The properts .filter_comments is deprecated and will "
                      "be removed in v1.0.", DeprecationWarning, 2)
        return list(self.values())

    def __str__(self) -> str:
        filters = [f"{self.name}/{key}" for key in self]
        outstr = (f"Filter system '{self.name}': {self.title}\n"
                  f"  spectral coverage: {', '.join(self.spectral_coverage)}\n"
                  f"  wave_unit: {self.read_kwargs['wave_unit']}\n"
                  f"  filters: {', '.join(filters)}")
        return outstr


class ExtCurvesLibrary(Library):
    """Library of extinction curves."""

    db_dir = "extinction_curves"
    aliases = {
        "items": "curves",
        "wave_col": "wave_column",
        "flux_col": "extinction_column",
        "flux_unit": "extinction_unit",
        }

    def __init__(self, name=None, curve_library=None):
        if curve_library is not None:
            warnings.warn("Constructing using the 'curve_library' argument is "
                          "deprecated and will raise an error in v1.0. Please "
                          "use the more general 'name' argument.",
                          DeprecationWarning, 2)
            name = curve_library
        if name is None and curve_library is None:
            raise ConstructorError("name must be passed to constructor")

        super().__init__(name)

    @property
    def curves(self):
        """Deprecated feature."""
        warnings.warn("The alias .curves is deprecated and will be "
                      "removed in v1.0. Please use .items() instead!",
                      DeprecationWarning, 2)
        return list(self.items())

    @property
    def curve_names(self):
        """Deprecated feature."""
        warnings.warn("The alias .curve_names is deprecated and will be "
                      "removed in v1.0. Please use the  more general "
                      ".keys() instead!", DeprecationWarning, 2)
        return list(self.keys())

    @property
    def curve_comments(self):
        """Deprecated feature."""
        warnings.warn("The properts .curve_comments is deprecated and will be "
                      "removed in v1.0.", DeprecationWarning, 2)
        return list(self.values())

    def __str__(self) -> str:
        outstr = (f"Extinction Curves '{self.name}': {self.title}\n"
                  f"  wave_unit: {self.read_kwargs['wave_unit']}\n"
                  f"  Templates: {self._items!s}")
        return outstr



# -*- coding: utf-8 -*-
from . import spextra
from . import database

from .spextra import Spextrum, Passband, ExtinctionCurve, get_vega_spectrum
from .database import spextra_database, DEFAULT_DATA
from .libraries import SpecLibrary, FilterSystem, ExtCurvesLibrary
from .configuration import config
from .exceptions import SpextraError



# -*- coding: utf-8 -*-
"""Tests for class Spextrum."""
from pathlib import Path

import pytest

import numpy as np
import astropy.units as u
from astropy.constants import c
from synphot import SpectralElement, SourceSpectrum, units

from spextra import Spextrum, Passband, spextra_database
from spextra.exceptions import SpextraError, NotInLibraryError

spextra_database.data_dir = Path(__file__).parent.parent.parent / 'database'
assert spextra_database.data_dir.exists()

# TODO: function scope might be better to isolate tests, but check performance
# offline seems to be about a 50% increase in runtime for function scope
@pytest.fixture(scope="class", name="spec")
def simple_spectrum():
    return Spextrum("kc96/s0")


# TODO: function scope might be better to isolate tests, but check performance
# offline seems to be about a 50% increase in runtime for function scope
@pytest.fixture(scope="class", name="bb_spec")
def black_body_spectrum():
    # Note: used to be Spextrum.black_body_spectrum(filter_curve="g") in a very
    #       old version...
    return Spextrum.black_body_spectrum(filter_curve="SLOAN/SDSS.z")


def test_make_passband_no_file():
    with pytest.raises(SpextraError) as e_info:
        _ = Passband.from_file(filename="blablabla")
        print(e_info)


class TestPassbandInstances:
    def test_alias(self):
        passband = Passband("U")
        assert isinstance(passband, SpectralElement)

    @pytest.mark.webtest
    def test_svo(self):
        """Test downloading from SVO."""
        # Use a file that is not stored in this repository.
        # passband = Passband("Paranal/HAWKI.Ks")
        passband = Passband("F110W")
        assert isinstance(passband, SpectralElement)

    def test_database(self):
        passband = Passband("elt/micado/Y")
        assert isinstance(passband, SpectralElement)

    def test_database_dash(self):
        passband = Passband("elt/micado/I-long")
        assert isinstance(passband, SpectralElement)

    def test_database_underscore(self):
        passband = Passband("elt/micado/H2_1-0S1")
        assert isinstance(passband, SpectralElement)

    @pytest.mark.usefixtures("mock_dir")
    def test_filename(self, mock_dir):
        passband = Passband.from_file(filename=mock_dir / "Y.dat")
        assert isinstance(passband, SpectralElement)


@pytest.mark.webtest
@pytest.mark.usefixtures("spec", "bb_spec")
class TestSpextrumInstances:
    """
    This class tests whether each method return the correct instances
    it also tests whether the methods are functional
    but it doesn't test for correct outputs see cls:TestSpextrum for that
    """

    def test_load(self, spec):
        assert isinstance(spec, Spextrum)

    def test_sub_class(self):
        assert issubclass(Spextrum, SourceSpectrum)

    def test_redshift(self, spec):
        sp2 = spec.redshift(z=1)
        assert isinstance(sp2, Spextrum)

    def test_add_emi_lines(self, spec):
        sp2 = spec.add_emi_lines([5000, 6000], [10, 20], [1e-15, 2e-15])
        assert isinstance(sp2, Spextrum)

    def test_add_abs_lines(self, spec):
        sp2 = spec.add_abs_lines([5000, 6000], [15, 20], [10, 12])
        assert isinstance(sp2, Spextrum)

    @pytest.mark.parametrize("unit", [u.mag, u.STmag, u.ABmag])
    @pytest.mark.parametrize("waves", [np.arange(1e4, 2.7e4, 100) * u.AA, None])
    def test_flat_spectrum(self, unit, waves):
        sp = Spextrum.flat_spectrum(
            waves=waves,
            amplitude=10*unit,
        )
        assert isinstance(sp, Spextrum)

    def test_mul_with_scalar(self, spec):
        sp2 = spec * 2
        assert isinstance(sp2, Spextrum)

    def test_sum_spectra(self, spec):
        sp2 = Spextrum("pickles/a0v")
        sp = spec + sp2
        assert isinstance(sp, Spextrum)

    def test_scale_to_magnitude(self, spec):
        sp2 = spec.scale_to_magnitude(amplitude=15*u.ABmag, filter_curve="g")
        assert isinstance(sp2, Spextrum)

    def test_rebin_spectra(self, spec):
        new_waves = np.linspace(np.min(spec.waveset),
                                np.max(spec.waveset),
                                100)
        sp2 = spec.rebin_spectra(new_waves=new_waves)
        assert isinstance(sp2, Spextrum)

    def test_get_magnitude(self):
        sp = Spextrum("pickles/a0v")
        mag = sp.get_magnitude("elt/micado/Y", system_name="AB")
        assert isinstance(mag, u.Quantity)

    @pytest.mark.webtest
    def test_black_body_spectrum(self, bb_spec):
        assert isinstance(bb_spec, Spextrum)

    @pytest.mark.webtest
    def test_black_body_spectrum_units(self):
        bb = Spextrum.black_body_spectrum(
                temperature=9500 * u.K,
                amplitude=0*u.ABmag,
                filter_curve="V",
                waves=np.arange(3e3, 8e3, 100)*u.AA,
        )
        assert isinstance(bb, Spextrum)
        bb = Spextrum.black_body_spectrum(
                temperature=9500*u.K,
                amplitude=0*u.ABmag,
                filter_curve="V",
                waves=None,
        )
        assert isinstance(bb, Spextrum)

    @pytest.mark.webtest
    def test_powerlaw_units(self):
        spec = Spextrum.powerlaw(
                x_0=9500*u.AA,
                amplitude=0*u.ABmag,
                filter_curve="V",
                waves=np.arange(3e3, 8e3, 100)*u.AA,
        )
        assert isinstance(spec, Spextrum)
        spec = Spextrum.powerlaw(
                x_0=9500*u.AA,
                amplitude=0*u.ABmag,
                filter_curve="V",
                waves=None,
        )
        assert isinstance(spec, Spextrum)

    @pytest.mark.webtest
    def test_photons_in_range(self, bb_spec):
        counts = bb_spec.photons_in_range(wmin=4000, wmax=5000)
        assert isinstance(counts, u.Quantity)

    @pytest.mark.webtest
    def test_smooth(self, bb_spec):
        # Note: The previous value throws am undersampling warning. Since this
        #       test is only about isinstance, just use a value that doesn't.
        #       This could maybe also be solved by passing a custom waves
        #       parameter to black_body_spectrum.
        # sp2 = sp.smooth(10*(u.m / u.s))
        sp2 = bb_spec.smooth(150*(u.km / u.s))
        assert isinstance(sp2, Spextrum)

    @pytest.mark.webtest
    def test_redden(self, bb_spec):
        sp2 = bb_spec.redden("calzetti/starburst", Ebv=0.1)
        assert isinstance(sp2, Spextrum)

    def test_deredden(self, bb_spec):
        sp2 = bb_spec.redden("gordon/smc_bar", Ebv=0.1)
        assert isinstance(sp2, Spextrum)

    def testing_nesting(self, spec):
        sp = spec.redshift(z=1).scale_to_magnitude(
            amplitude=15*u.ABmag, filter_curve="g").redden(
                "calzetti/starburst", Ebv=0.1)
        assert isinstance(sp, Spextrum)


@pytest.mark.usefixtures("spec")
class TestSpextrum:

    def test_wrong_load(self):
        with pytest.raises(NotInLibraryError):
            _ = Spextrum("kc96/wrong_name")

    @pytest.mark.parametrize("unit", [u.mag, u.STmag, u.ABmag])
    def test_ref_spectrum_is_right(self, unit):
        sp1 = Spextrum.flat_spectrum(amplitude=10*unit)
        sp2 = Spextrum.flat_spectrum(amplitude=11*unit)
        if unit == u.mag:
            flux1 = sp1(sp1.waveset[(sp1.waveset.value > 7000 - 200) &
                                    (sp1.waveset.value < 7000 + 200)]).value
            flux2 = sp2(sp2.waveset[(sp2.waveset.value > 7000 - 200) &
                                    (sp2.waveset.value < 7000 + 200)]).value

        else:
            waves = np.arange(1000, 1e4, 1) * u.AA
            flux1 = sp1(waves)
            flux2 = sp2(waves)

        mean = np.mean(flux1 / flux2)
        assert np.isclose(mean, 10**0.4)

    @pytest.mark.webtest
    @pytest.mark.parametrize("unit", [u.mag, u.ABmag, u.STmag])
    def test_correct_scaling(self, unit, spec):
        sp1 = spec.scale_to_magnitude(
            amplitude=14*unit, filter_curve="SLOAN/SDSS.rprime_filter")
        sp2 = spec.scale_to_magnitude(
            amplitude=15*unit, filter_curve="SLOAN/SDSS.rprime_filter")

        flux1 = sp1(sp1.waveset[(sp1.waveset.value > 6231 - 200) &
                                (sp1.waveset.value < 6231 + 200)]).value
        flux2 = sp2(sp2.waveset[(sp2.waveset.value > 6231 - 200) &
                                (sp2.waveset.value < 6231 + 200)]).value

        mean = np.mean(flux1 / flux2)
        assert np.isclose(mean, 10**0.4)

    @pytest.mark.webtest
    @pytest.mark.parametrize("filt", ["U", "B", "V", "R", "I", "J", "H", "Ks"])
    def test_vega2ab(self, filt):
        """
        test if the convertion between AB and Vega is correct
        conversions taken from:
        http://www.astronomy.ohio-state.edu/~martini/usefuldata.html

        absolute tolerance set to 0.15 mag to account for filter differences

        Parameters
        ----------
        filt: str
            name of the filter
        """
        ab2vega = {"U":  0.79,   # magAB - magVega taken from
                   "B": -0.09,   #
                   "V":  0.02,
                   "R":  0.21,
                   "I":  0.45,
                   "J":  0.91,
                   "H":  1.39,
                   "Ks": 1.85}

        sp = Spextrum.flat_spectrum(amplitude=0*u.ABmag)

        magAB = sp.get_magnitude(filt, system_name="AB")
        magVega = sp.get_magnitude(filt, system_name="vega")

        diff = (magAB.value - magVega.value)

        assert np.isclose(diff, ab2vega[filt], atol=0.15)

    @pytest.mark.webtest
    @pytest.mark.parametrize("spec_name", ["vega", "sun", "brown/NGC5992"])
    @pytest.mark.parametrize("filters", [["2MASS/2MASS.Ks", "elt/micado/Ks"],
                                         ["J", "elt/micado/J"],
                                         # ["Generic/Bessell.B", "B"],
                                         ])
    def test_magnitudes(self, spec_name, filters):
        sp = Spextrum(spec_name)
        filter1 = filters[0]
        filter2 = filters[1]
        mag1 = sp.get_magnitude(filter_curve=filter1)
        mag2 = sp.get_magnitude(filter_curve=filter2)

        assert np.isclose(mag1.value, mag2.value, atol=0.3)

    def test_hz2angstrom(self):

        waves = np.array([1, 2, 3]) * u.Hz
        flux = np.array([1, 1, 2]) * units.FLAM

        sp = Spextrum.from_arrays(waves, flux)

        inwaves = c.value / waves.value
        outwaves = sp.waveset.to(u.m).value

        assert np.isclose(inwaves[::-1], outwaves).all()

    def test_spectrum_cut(self, spec):
        w1 = 3001*u.AA
        w2 = 4000*u.AA
        sp2 = spec.cut(w1, w2)
        assert np.isclose(sp2.wave_min.value, w1.value,
                          atol=np.abs(sp2.waveset[0].value -
                                      sp2.waveset[1].value))



# -*- coding: utf-8 -*-
"""Holds the Config class and an instance to use as global configuration."""

from pathlib import Path
from dataclasses import dataclass, asdict
import yaml

from .exceptions import SpextraError

__all__ = ["config"]
__data_dir__ = Path(__file__).parent / "data"
__config_file__ = __data_dir__ / "config.yml"


@dataclass
class Config:
    """
    Set up the configuration for the database.

    This class reads a yaml file in the directory data_dir that contains the
    default configuration of the database that contain default values for the
    following attributes:

    database_url: The location of the remote database
    cache_dir: The location of the database in the local hard disk, where the
        files are saved. Default is Null (None) which then lead to
        .spextra_cache
    retry: The number of times to re-attempt a download.

    Normally, the user will not need to call this class, except when chaning

    If the user needs to change these values the class can be simply called as

    >> Config(cache_dir="path_to_new_data_directory")

    and the new path will be used from that moment on
    """

    database_url: str
    cache_dir: Path
    retry: int
    registry_file: Path

    def __post_init__(self):
        self.cache_dir = Path(self.cache_dir)
        self.registry_file = Path(self.registry_file)

        if not self.cache_dir.is_absolute():
            self.cache_dir = Path.home() / self.cache_dir

        if not self.registry_file.is_absolute():
            self.registry_file = __data_dir__ / self.registry_file

        if not self.registry_file.exists():
            raise SpextraError("registry file not found")

        # TODO: dump back into config file
        # TODO: add method to restore defaults

    @classmethod
    def from_yaml(cls, path: Path):
        with path.open(encoding="utf-8") as file:
            return cls(**yaml.safe_load(file))

    def __str__(self) -> str:
        sout = "SpeXtra Configuration:\n"
        for k, v in asdict(config).items():
            sout += f"  {k}: {v!s}\n"
        return sout


config = Config.from_yaml(__config_file__)



# -*- coding: utf-8 -*-
"""speXtra: A python tool to manage and manipulate astronomical spectra."""

import numbers
import warnings
from pathlib import Path

import numpy as np
from scipy.ndimage import gaussian_filter1d
from more_itertools import always_iterable

import astropy.units as u
from astropy.constants import c as speed_of_light
from astropy.modeling.models import Scale

from synphot import (units, SourceSpectrum, SpectralElement, Observation,
                     BaseUnitlessSpectrum, ReddeningLaw, utils)
from synphot.models import (Empirical1D, GaussianFlux1D, Gaussian1D, Box1D,
                            ConstFlux1D, BlackBody1D, PowerLawFlux1D)
from synphot.specio import read_ascii_spec, read_fits_spec
from synphot import exceptions

from .database import DEFAULT_DATA
from .containers import SpectrumContainer, FilterContainer, ExtCurveContainer
from .downloads import download_svo_filter
from .exceptions import SpextraError, ArgumentError, ConstructorError


__all__ = ["Spextrum", "Passband", "ExtinctionCurve", "get_vega_spectrum"]


class Passband(SpectralElement, FilterContainer):
    """
    Class to handle astronomical filters.

    TODO: Implement proper __add__, __sub__, __mul__, etc

    """

    DEFAULT_FILTERS = DEFAULT_DATA.filters

    def __init__(self, filter_name=None, modelclass=None, **kwargs):
        if filter_name is not None:
            self.from_filter_name(filter_name)
        elif modelclass is not None:
            self.from_modelclass(modelclass, **kwargs)
        else:
            raise ConstructorError(
                "Either filter_name or modelclass must be passed."
            )

    # @classmethod
    def from_filter_name(self, filter_name: str):
        filter_name = self.DEFAULT_FILTERS.get(filter_name, filter_name)

        FilterContainer.__init__(self, filter_name)

        if self.is_in_library and self.library.is_in_database:
            self._database_loader()
        else:
            try:  # try to download it from SVO
                self._svo_loader(filter_name)
            except ValueError:
                raise SpextraError(f"Filter {filter_name} not found.")

    # @classmethod
    def from_modelclass(self, modelclass, **kwargs):
        SpectralElement.__init__(self, modelclass, **kwargs)

    def _database_loader(self):
        """
        Load a filter from the database.

        Returns
        -------
        meta: metadata of the spectra (header)
        lam: wavelengths
        flux: flux
        """
        meta, wave, flux = _read_spec(self.path, self.library.data_type,
                                      **self.library.read_kwargs)

        # TODO: why?
        wave = wave.to(u.AA)

        SpectralElement.__init__(self, Empirical1D, points=wave,
                                 lookup_table=flux.value, meta=meta)

    def _svo_loader(self, filter_name):
        wave, trans = download_svo_filter(filter_name)
        meta = {"filter_name": filter_name, "source": "SVO"}
        self.library.read_kwargs["wave_unit"] = u.Unit(wave.unit)

        SpectralElement.__init__(self, Empirical1D, points=wave,
                                 lookup_table=trans, meta=meta)

    @classmethod
    def from_vectors(
            cls,
            waves,
            trans,
            meta: dict | None = None,
            wave_unit: u.Unit = u.AA):
        """
        Create a ``Passband`` directly from from array-like.

        Parameters
        ----------
        waves: list-like
        trans: list-like
        meta: dictionary containing the metadata
        wave_unit: u.Quantity, defaulted to angstroms

        Returns
        -------
            New ``Passband`` instance.
        """
        waves <<= wave_unit
        modelclass = SpectralElement(
            Empirical1D, points=waves, lookup_table=trans, meta=meta
        )
        return cls(modelclass=modelclass)

    @classmethod
    def from_file(cls, filename: Path | str, **kwargs):
        """
        Create a ``Passband`` from a file.

        Parameters
        ----------
        filename : TYPE
            DESCRIPTION.

        Returns
        -------
            New ``Passband`` instance.
        """
        try:
            modelclass = SpectralElement.from_file(str(filename), **kwargs)
        except FileNotFoundError as err:
            raise SpextraError(f"File {filename} not found.") from err

        return cls(modelclass=modelclass)

    @classmethod
    @u.quantity_input(equivalencies=u.spectral())
    def gaussian(
            cls,
            center: u.Quantity[u.AA],
            fwhm: u.Quantity[u.AA],
            peak,
            **kwargs):
        """
        Create a ``Passband`` with a gaussian shape with given user parameters.

        Parameters
        ----------
        center : TYPE
            DESCRIPTION.
        fwhm : TYPE
            DESCRIPTION.
        peak : TYPE
            DESCRIPTION.

        Returns
        -------
            New ``Passband`` instance.
        """
        sigma = fwhm.to(u.AA) / (2.0 * np.sqrt(2.0 * np.log(2.0)))

        modelclass = SpectralElement(
            Gaussian1D,
            amplitude=peak,
            mean=center.to(u.AA),
            stddev=sigma,
            **kwargs,
        )
        return cls(modelclass=modelclass)

    @classmethod
    @u.quantity_input(equivalencies=u.spectral())
    def square(
            cls,
            wmin: u.Quantity[u.AA],
            wmax: u.Quantity[u.AA],
            transmission):
        """
        Create a ``Passband`` with a rectangular shape.

        Parameters
        ----------
        wmin : TYPE
            DESCRIPTION.
        wmax : TYPE
            DESCRIPTION.
        transmission : TYPE
            DESCRIPTION.

        Returns
        -------
            New ``Passband`` instance.
        """
        center = (wmax + wmin) * 0.5
        width = wmax - wmin
        modelclass = SpectralElement(
            Box1D,
            amplitude=transmission,
            x_0=center.to(u.AA),
            width=width.to(u.AA),
        )
        return cls(modelclass=modelclass)


class ExtinctionCurve(ReddeningLaw, ExtCurveContainer):
    """
    Class to handle extinction curves.

    TODO: Implement proper __add__, __sub__, __mul__, etc

    """

    DEFAULT_CURVES = DEFAULT_DATA.extcurves

    def __init__(self, curve_name: str | None = None, modelclass=None):
        if curve_name is not None:
            self.from_curve_name(curve_name)
        elif modelclass is not None:
            self.from_modelclass(modelclass)
        else:
            raise ConstructorError(
                "Either curve_name or modelclass must be passed."
            )

    def from_curve_name(self, curve_name: str):
        curve_name = self.DEFAULT_CURVES.get(curve_name, curve_name)

        ExtCurveContainer.__init__(self, curve_name)
        self._database_loader()

    def from_modelclass(self, modelclass):
        ReddeningLaw.__init__(self, modelclass)

    def _database_loader(self):
        """
        Load a filter from the database.

        Returns
        -------
        meta: metadata of the spectra (header)
        lam: wavelengths
        flux: flux
        """
        meta, wave, flux = _read_spec(self.path, self.library.data_type,
                                      **self.library.read_kwargs)

        ReddeningLaw.__init__(self, Empirical1D, points=wave,
                              lookup_table=flux, meta=meta)

    @classmethod
    def from_vectors(
            cls,
            waves,
            flux,
            meta: dict | None = None,
            wave_unit: u.Unit = u.AA):
        """
        Create a ``Passband`` directly from from array-like.

        Parameters
        ----------
        waves: list-like
        flux: list-like
        meta: dictionary containing the metadata
        wave_unit: u.Quantity, defaulted to angstroms
        flux_unit: u.Quantiy, defaulted to FLAM

        Returns
        -------
        Passband
        """
        waves <<= wave_unit

        modelclass = SpectralElement(
            Empirical1D, points=waves, lookup_table=flux, meta=meta
        )
        return cls(modelclass=modelclass)

    @classmethod
    def from_file(cls, filename: Path | str, **kwargs):
        modelclass = ReddeningLaw.from_file(filename, **kwargs)
        return cls(modelclass=modelclass)


class Spextrum(SourceSpectrum, SpectrumContainer):
    """
    Class to handle spectra.

    This class download, load, stores and manipulates the spectra.

    This class can be initialized with a remote file which will be downloaded
    from the database or with a ``synphot.BaseSpectrum`` or ``SourceSpectrum``.


    Parameters
    ----------
    template_name : Name of the template to download with format
    library/template e.g. "kc96/s0"
    modelclass : SourceSpectrum or BaseSpectrum

    """

    DEFAULT_SPECTRA = DEFAULT_DATA.spectra

    def __init__(self, template_name=None, modelclass=None, **kwargs):
        if template_name is not None:
            self.from_template_name(template_name, **kwargs)
        elif modelclass is not None:
            self.from_modelclass(modelclass, **kwargs)
        else:
            raise ValueError(
                "Either template_name or modelclass must be passed."
            )

    def from_template_name(self, template_name: str, **kwargs):
        template_name = self.DEFAULT_SPECTRA.get(template_name, template_name)

        SpectrumContainer.__init__(self, template_name)
        self._database_loader(**kwargs)
        self.repr = f"({self.template})"

    def from_modelclass(self, modelclass, **kwargs):
        SourceSpectrum.__init__(self, modelclass=modelclass, **kwargs)
        # FIXME: this None is problematic, as it's used for the repr, which
        #        makes an instance of this look like None o.O
        self.repr = None

    def _database_loader(self, **kwargs):
        """
        Load a template from the database.

        Returns
        -------
        meta: metadata of the spectra (header)
        lam: wavelengths
        flux: flux
        """
        meta, wave, flux = _read_spec(
            self.path, self.library.data_type, **self.library.read_kwargs
        )

        SourceSpectrum.__init__(
            self,
            Empirical1D,
            points=wave,
            lookup_table=flux,
            meta=meta,
            **kwargs,
        )

    @classmethod
    def from_arrays(
        cls,
        waves,
        flux,
        meta: dict | None = None,
        wave_unit: u.Unit = u.AA,
        flux_unit: u.Unit = units.FLAM,
    ):
        """
        Create a ``Passband`` directly from from an array-like.

        Parameters
        ----------
        waves: array-like
        flux: array-like
        meta: dictionary containing the metadata
        wave_unit: u.Quantity, defaulted to angstroms
        flux_unit: u.Quantiy, defaulted to FLAM

        Returns
        -------
        Passband
        """
        try:
            waves <<= wave_unit
        except u.UnitConversionError:
            waves = waves.to(wave_unit, equivalencies=u.spectral())
        flux <<= flux_unit

        modelclass = SourceSpectrum(
            Empirical1D, points=waves, lookup_table=flux, meta=meta
        )

        spex = cls(modelclass=modelclass)
        spex.repr = repr(spex.model)
        return spex

    @classmethod
    def from_file(cls, filename: Path | str, **kwargs):
        """
        Call ``synphot.SourceSpectrum.from_file()`` method.

        Parameters
        ----------
        filename
        kwargs

        Returns
        -------
        spex : Spextrum
            New ``Spextrum`` instance.
        """
        modelclass = SourceSpectrum.from_file(filename, **kwargs)
        spex = cls(modelclass=modelclass)
        spex.repr = f".from_file({filename})"

        return spex

    @classmethod
    @u.quantity_input
    def flat_spectrum(
        cls,
        amplitude: u.Quantity[u.ABmag] = 0 * u.ABmag,
        waves: u.Quantity[u.AA] | None = None,
    ):
        """
        Create a flat spectrum in the preferred system scaled to a magnitude.

        Defaults to a zero magnitude spectrum.

        Parameters
        ----------
        amplitude: float, u.Quantity
            amplitude/magnitude of the reference spectrum, default=0
            default is u.ABmag
            for vega use u.mag

        waves: The waveset of the reference spectrum if not Vega
           if not provided they will be created at a resolution of R~800

        Returns
        -------
        spex : Spextrum
            New ``Spextrum`` instance.
        """
        if waves is None:
            waves = _default_waves()
        waves = waves.to(u.AA).value

        # The if-statement below also allowed amplitude.unit to be
        # u.Unit("vegamag"). Vegamag is removed from astropy, so the
        # if-statement is removed too. This assertion is added as a sanity
        # check ensure that any bugs introduced by changing the if-statement
        # is explicitly raised here, instead of silently propagated.
        assert (
            not amplitude.unit.to_string() == "vegamag"
        ), "The use of vegamag is deprecated."
        if amplitude.unit is u.Unit("mag"):
            spex = get_vega_spectrum()
            spex *= 10 ** (-0.4 * amplitude.value)
            # system_name = amplitude.unit
        else:
            const = ConstFlux1D(amplitude=amplitude)
            spex = cls(
                modelclass=Empirical1D, points=waves, lookup_table=const(waves)
            )
            # system_name = amplitude.unit

        spex.repr = f".flat_spectrum(amplitude={amplitude!r})"
        return spex

    @classmethod
    @u.quantity_input
    def black_body_spectrum(
        cls,
        temperature: u.Quantity[u.K] = 9500 * u.K,
        amplitude: u.Quantity[u.ABmag] = 0 * u.ABmag,
        filter_curve: str | Passband | SpectralElement | None = None,
        waves: u.Quantity[u.AA] | None = None,
    ):
        """
        Produce a blackbody spectrum for a given temperature.

        And scale it to a magnitude in a filter.

        Parameters
        ----------
        temperature: the temperature in Kelvin degrees
        amplitude: ``astropy.Quantity``, float
            The value that the spectrum should have in the given filter.
            Acceptable astropy quantities are:

            - u.mag : Vega magnitudes
            - u.ABmag : AB magnitudes
            - u.STmag : HST magnitudes
            - u.Jy : Jansky per filter bandpass

            Additionally the ``FLAM`` and ``FNU`` units from ``synphot.units``
            can be used when passing the quantity for `amplitude`.

        filter_curve : str
            Name of a filter from

            - a generic filter name (see ``FILTER_DEFAULTS``)
            - a spanish-vo filter service ref. (e.g. ``"Paranal/HAWKI.Ks"``)
            - a filter in the spextra database
            - a filename with the filter file
            - a ``Passband`` or ``synphot.SpectralElement`` object

        Returns
        -------
        a scaled black-body spectrum
        """
        if waves is None:
            waves = _default_waves()
        waves = waves.to(u.AA)
        blackbody = BlackBody1D(temperature=temperature.to(u.K).value)

        spex = cls(modelclass=Empirical1D, points=waves,
                   lookup_table=blackbody(waves))
        spex = spex.scale_to_magnitude(amplitude=amplitude,
                                       filter_curve=filter_curve)
        spex.repr = (f".black_body_spectrum(amplitude={amplitude!r}, "
                     f"temperature={temperature!r}, "
                     f"filter_curve={filter_curve!r})")
        return spex

    @classmethod
    @u.quantity_input(equivalencies=u.spectral())
    def powerlaw(
        cls,
        alpha=1,
        x_0: u.Quantity[u.AA] = 5000 * u.AA,
        amplitude: u.Quantity[u.ABmag] = 0 * u.ABmag,
        filter_curve: str | Passband | SpectralElement | None = None,
        waves: u.Quantity[u.AA] | None = None,
    ):
        """
        Return a power law spectrum.

        F(lambda) ~ lambda^alpha scaled to a magnitude (amplitude) in a
        particular band.

        Parameters
        ----------
        alpha : The spectral slope
        x_0 : float, u.Quantity
           Pivot wavelength
        amplitude : float, u.Quantity
            normalize the spectrum to that quantity
        filter_curve : str
            Name of a filter from

            - a generic filter name (see ``FILTER_DEFAULTS``)
            - a spanish-vo filter service ref. (e.g. ``"Paranal/HAWKI.Ks"``)
            - a filter in the spextra database
            - a filename with the filter file
            - a ``Passband`` or ``synphot.SpectralElement`` object

        Returns
        -------
        spex : Spextrum
            New ``Spextrum`` instance.
        """
        if waves is None:
            waves = _default_waves()
        waves = waves.to(u.AA)
        power = SourceSpectrum(
            PowerLawFlux1D, amplitude=1, x_0=x_0, alpha=alpha)
        spex = cls(modelclass=Empirical1D, points=waves,
                   lookup_table=power(waves))

        spex = spex.scale_to_magnitude(
            amplitude=amplitude, filter_curve=filter_curve)

        spex.repr = (f"Spextrum.powerlaw(alpha={alpha!r}, x_0={x_0!r}, "
                     f"amplitude={amplitude!r}, filter_curve={filter_curve!r})")
        return spex

    @classmethod
    def emission_line_spectra(cls, *args, **kwargs):
        warnings.warn(
            "The 'emission_line_spectra' constructor is deprecated and will be"
            " removed in a future version. Please use the identical "
            "'emission_line_spectrum' instead.", DeprecationWarning,
            stacklevel=2)
        return cls.emission_line_spectrum(*args, **kwargs)

    @classmethod
    @u.quantity_input(equivalencies=u.spectral())
    def emission_line_spectrum(
        cls,
        center: u.Quantity[u.AA],
        fwhm: u.Quantity[u.AA],
        flux,
        amplitude: u.Quantity[u.ABmag] = 40 * u.ABmag,
        waves: u.Quantity[u.AA] | None = None,
    ):
        """
        Create a emission line spextrum superimpossed to a faint continuum.

        Parameters
        ----------
        center
        fwhm
        flux
        amplitude
        waves

        Returns
        -------
        spex : Spextrum
            New ``Spextrum`` instance.
        """
        if waves is None:
            wmin = np.min(center.to(u.AA).value) - 2000
            wmax = np.max(center.to(u.AA).value) + 2000
            step = np.min(fwhm.to(u.AA).value) / 3  # for Nyquist sampling
            waves = np.arange(wmin, wmax, step) * u.AA

        spex = cls.flat_spectrum(amplitude=amplitude, waves=waves)
        spex = spex.add_emi_lines(center=center, fwhm=fwhm, flux=flux)

        return spex

    @property
    def wave_min(self):
        return np.min(self.waveset)

    @property
    def wave_max(self):
        return np.max(self.waveset)

    def cut(self, wave_min, wave_max):
        """
        Cut the spectrum between wmin and wmax.

        Parameters
        ----------
        wave_min: float, u.Quantity,
        wave_max: float, u.Quantity,

        Returns
        -------
        spex : Spextrum
            New ``Spextrum`` instance.
        """
        wave_min <<= u.AA
        wave_max <<= u.AA

        new_waves = self.waveset[
            (self.waveset >= wave_min) & (self.waveset <= wave_max)
        ]
        spex = Spextrum(
            modelclass=Empirical1D,
            points=new_waves,
            lookup_table=self(new_waves),
            meta=self.meta,
        )
        spex = self._restore_attr(spex)

        return spex

    def redshift(self, z=0, vel=0):
        """
        Red- or blueshift a spectrum.

        Parameters
        ----------
        z : TYPE, optional
            Redshift. The default is 0.
        vel : TYPE, optional
            Radial velocity. If no unit are present it is assumed to be in m/s.
            The default is 0.

        Raises
        ------
        ValueError
            Raised if `z` or `vel` are unphysical.

        Returns
        -------
        spex : Spextrum
            New ``Spextrum`` instance.
        """
        if vel != 0:
            vel <<= u.m / u.s  # assumed to be in m/s

            z = (vel / speed_of_light).value
        if z <= -1:
            raise ValueError("Redshift or velocity unphysical")

        lam = self.waveset * (1 + z)
        flux = self(self.waveset)
        self.meta.update({"redshift": z})

        spex = Spextrum(
            modelclass=Empirical1D,
            points=lam,
            lookup_table=flux,
            meta=self.meta,
        )
        spex = self._restore_attr(spex)
        return spex

    @u.quantity_input
    def scale_to_magnitude(
        self,
        amplitude: u.Quantity[u.ABmag],
        filter_curve: str | Passband | SpectralElement | None = None,
    ):
        """
        Scale a Spectrum to a value in a filter.

        Copied from scopesim.effects.ter_curves.scale_spectrum with slight
        modifications.

        Parameters
        ----------
        amplitude : ``astropy.Quantity``, float
            The value that the spectrum should have in the given filter.
            Acceptable astropy quantities are:

            - u.mag : Vega magnitudes
            - u.ABmag : AB magnitudes
            - u.STmag : HST magnitudes
            - u.Jy : Jansky per filter bandpass

            Additionally the ``FLAM`` and ``FNU`` units from ``synphot.units``
            can be used when passing the quantity for `amplitude`:

        filter_curve : str
            Name of a filter from

            - a generic filter name (see ``DEFAULT_FILTERS``)
            - a spanish-vo filter service ref. (e.g. ``"Paranal/HAWKI.Ks"``)
            - a filter in the spextra database
            - the path to a filename containing the curve (see ``Passband``)
            - a ``Passband`` or ``synphot.SpectralElement`` instance

        filter_file: str
            A file with a transmission curve


        Returns
        -------
        spectrum : a Spectrum
            Input spectrum scaled to the given amplitude in the given filter
        """
        if Path(filter_curve).exists():
            filter_curve = Passband.from_file(filename=filter_curve)
        elif not isinstance(filter_curve, (Passband, SpectralElement)):
            filter_curve = Passband(filter_curve)

        ref_spec = self.flat_spectrum(amplitude=amplitude)

        ref_flux = Observation(ref_spec,
                               filter_curve).effstim(flux_unit=units.PHOTLAM)
        real_flux = Observation(self,
                                filter_curve).effstim(flux_unit=units.PHOTLAM)

        scale_factor = ref_flux / real_flux
        spex = self * scale_factor
        spex = self._restore_attr(spex)
        spex.meta.update({"magnitude": amplitude,
                          "filter_curve": filter_curve})

        return spex

    def get_magnitude(
        self,
        filter_curve: str | Passband | SpectralElement | None = None,
        system_name: str = "AB",
    ):
        """
        Obtain the magnitude in filter for a user specified photometric system.

        Parameters
        ----------
        filter_curve : str
            Name of a filter from

            - a generic filter name (see ``DEFAULT_FILTERS``)
            - a spanish-vo filter service ref. (e.g. ``"Paranal/HAWKI.Ks"``)
            - a filter in the spextra database
            - the path to the file containing the filter (see ``Passband``)
            - a ``Passband`` or ``synphot.SpectralElement`` object

        system_name: str
            The photometric system Vega, AB or ST


        Returns
        -------
        spectrum : a Spextrum
            Input spectrum scaled to the given amplitude in the given filter.
        """
        if Path(filter_curve).exists():
            filter_curve = Passband.from_file(filename=filter_curve)
        elif not isinstance(filter_curve, (Passband, SpectralElement)):
            filter_curve = Passband(filter_curve)

        if system_name.lower() in ["vega"]:
            unit = u.mag
        elif system_name.lower() in ["st", "hst"]:
            unit = u.STmag
        else:
            unit = u.ABmag

        ref_spec = self.flat_spectrum(amplitude=0 * unit)
        ref_flux = Observation(ref_spec, filter_curve).effstim(
            flux_unit=units.PHOTLAM
        )
        real_flux = Observation(self, filter_curve).effstim(
            flux_unit=units.PHOTLAM
        )

        mag = -2.5 * np.log10(real_flux.value / ref_flux.value)

        return mag * unit

    def photons_in_range(
        self,
        wmin=None,
        wmax=None,
        area=1 * u.cm**2,
        filter_curve: str | Passband | SpectralElement | None = None,
    ):
        """
        Return the number of photons between wave_min and wave_max.

        Or within a bandpass (filter).

        Parameters
        ----------
        wmin :
            [Angstrom]
        wmax :
            [Angstrom]
        area : u.Quantity
            [cm2]
        filter_curve : str
            Name of a filter from

            - a generic filter name (see ``DEFAULT_FILTERS``)
            - a spanish-vo filter service ref. (e.g. ``"Paranal/HAWKI.Ks"``)
            - a filter in the spextra database
            - the path to the file containing the filter (see ``Passband``)
            - a ``Passband`` or ``synphot.SpectralElement`` object

        Returns
        -------
        counts : u.Quantity array

        """
        area <<= u.cm**2
        wmin = _angstrom_value(wmin)
        wmax = _angstrom_value(wmax)

        if filter_curve is None:
            # this makes a bandpass out of wmin and wmax
            try:
                mid_point = 0.5 * (wmin + wmax)
                width = abs(wmax - wmin)
                filter_curve = SpectralElement(
                    Box1D, amplitude=1, x_0=mid_point, width=width
                )
            except ValueError:
                raise ArgumentError("Please specify wmin/wmax or a filter")

        elif Path(filter_curve).exists():
            filter_curve = Passband.from_file(filename=filter_curve)
        elif not isinstance(filter_curve, (Passband, SpectralElement)):
            filter_curve = Passband(filter_curve)

        obs = Observation(self, filter_curve)
        counts = obs.countrate(area=area)

        return counts

    def get_flux(
        self,
        wmin=None,
        wmax=None,
        filter_curve: str | Passband | SpectralElement | None = None,
        flux_unit: u.Unit = units.FLAM,
    ):
        """
        Return the flux within a passband.

        Parameters
        ----------
        wmin : float, u.Quantity
            minimum wavelength
        wmax : float, u.Quantity
            maximum wavelength
        filter_curve : str
            Name of a filter from

            - a generic filter name (see ``DEFAULT_FILTERS``)
            - a spanish-vo filter service ref. (e.g. ``"Paranal/HAWKI.Ks"``)
            - a filter in the spextra database
            - the path to the file containing the filter (see ``Passband``)
            - a ``Passband`` or ``synphot.SpectralElement`` object

        flux_unit: synphot.units, u.Quantity

        Returns
        -------
        flux

        """
        wmin = _angstrom_value(wmin)
        wmax = _angstrom_value(wmax)

        if filter_curve is None:
            # this makes a bandpass out of wmin and wmax
            try:
                filter_curve = Passband.square(wmin=wmin, wmax=wmax)
            except ValueError:
                raise ArgumentError("Please specify wmin/wmax or a filter")
        elif Path(filter_curve).exists():
            filter_curve = Passband.from_file(filename=filter_curve)
        elif not isinstance(filter_curve, (Passband, SpectralElement)):
            filter_curve = Passband(filter_curve)

        flux = Observation(self, filter_curve).effstim(flux_unit=flux_unit)

        return flux

    def add_emi_lines(self, center, fwhm, flux):
        """
        Add emission lines to an `Spextrum`.

        TODO: accept different profiles (Lorentz1D, Voigt1D, etc)

        Parameters
        ----------
        center: float, list, np.ndarray, u.Quantity
            The center of the line

        fwhm:  float, list, np.ndarray, u.Quantity
            The FWHM of the line
        flux: float, list, np.ndarray, u.Quantity
            The Equivalent Width of the line

        Returns
        -------
        Spextrum

        """
        # Converting flux is done by GaussianFlux1D anyway...
        centers = list(always_iterable(_angstrom_value(center)))
        fluxes = list(always_iterable(flux))
        fwhms = list(always_iterable(_angstrom_value(fwhm)))

        self.meta.update({"em_lines": {"center": centers,
                                       "flux": fluxes,
                                       "fwhm": fwhms}})

        for cen, flx, fwh in zip(centers, fluxes, fwhms):
            line = GaussianFlux1D(mean=cen, total_flux=flx, fwhm=fwh)
            # lam = line.sampleset(factor_step=0.3)  # bit better than Nyquist
            lam = self.waveset.to(u.AA).value
            g_em = SourceSpectrum(Empirical1D, points=lam,
                                  lookup_table=line(lam))
            self += g_em

        spex = self._restore_attr(Spextrum(modelclass=self))
        return spex

    def add_abs_lines(self, center, ew, fwhm):
        """
        Add an absorption line to a spectrum.

        With center, fwhm and equivalent width specified by the user.
        It also supports emission lines if `ew` is negative.

        TODO: accept different profiles (Lorentz1D, Voigt1D, etc)

        Parameters
        ----------
        center: float, list, np.ndarray, u.Quantity
            The center of the line
        fwhm:  float, list, np.ndarray, u.Quantity
            The FWHM of the line
        ew: float, list, np.ndarray, u.Quantity
            The Equivalent Width of the line

        Returns
        -------
        Spextrum
        """
        centers = list(always_iterable(_angstrom_value(center)))
        eqws = list(always_iterable(_angstrom_value(ew)))
        fwhms = list(always_iterable(_angstrom_value(fwhm)))

        self.meta.update({"em_lines": {"center": centers,
                                       "ew": eqws,
                                       "fwhm": fwhms}})

        for cent, eqw, fwh in zip(centers, eqws, fwhms):
            # to keep the convention that EL are negative and ABS are positive
            sign = -1 * np.sign(eqw)
            left, right = cent - np.abs(eqw / 2), cent + np.abs(eqw / 2)
            wavelengths = self.waveset[(self.waveset.value >= left) &
                                       (self.waveset.value <= right)]
            fluxes = units.convert_flux(wavelengths=wavelengths,
                                        fluxes=self(wavelengths),
                                        out_flux_unit=units.FLAM)
            flux = np.trapz(fluxes.value, wavelengths.value)
            line = GaussianFlux1D(mean=cent, total_flux=sign * flux,
                                  fwhm=fwh)
            lam = line.sampleset(factor_step=0.35)  # bit better than Nyquist
            flux = line(lam).clip(min=0.0)
            g_abs = SourceSpectrum(Empirical1D, points=lam,
                                   lookup_table=flux)

            self += g_abs

            if (self(wavelengths).value < 0).any():
                warnings.warn("Warning: Flux<0 for specified EW and FHWM, "
                              "setting it to Zero")
                waves = self.waveset[self(self.waveset) < 0]
                zero_sp = SourceSpectrum(Empirical1D, points=waves,
                                         lookup_table=-1 * self(waves).value)
                self += zero_sp

        spex = self._restore_attr(Spextrum(modelclass=self))
        return spex

    @staticmethod
    def _get_extinct(curve_name: str, Ebv, Av, Rv):
        if Av is not None:
            Ebv = Av / Rv
        extcurve = ExtinctionCurve(curve_name)
        return extcurve.extinction_curve(Ebv)

    def redden(
        self,
        curve_name: str,
        Ebv: float = 0.0,
        Av: float | None = None,
        Rv: float = 3.1,
    ):
        """
        Attenuate a spectrum with a extinction curve normalized to a E(B-V).

        Parameters
        ----------
        curve_name: str, name of the extinction curve
        Av: float, Av parameter
        Ebv: float, E(B-V) color excess

        Returns
        -------
        spex : Spextrum
            New attenuated ``Spextrum`` instance.
        """
        extinction = self._get_extinct(curve_name, Ebv, Av, Rv)
        spex = Spextrum(modelclass=self.model * extinction.model)
        spex = self._restore_attr(spex)
        spex.meta.update({"Reddening with": curve_name,
                          "Reddening with E(B-V)": Ebv})
        return spex

    def deredden(
        self,
        curve_name: str,
        Av: float | None = None,
        Ebv: float = 0.0,
        Rv: float = 3.1,
    ):
        """
        De-redden a spectrum.

        Parameters
        ----------
        curve_name: str, name of the extinction curve
        Av: float, Av parameter
        Ebv: float, E(B-V) color excess

        Returns
        -------
        spex : Spextrum
            New attenuated ``Spextrum`` instance.
        """
        extinction = self._get_extinct(curve_name, Ebv, Av, Rv)
        spex = Spextrum(modelclass=self.model / extinction.model)
        spex = self._restore_attr(spex)
        spex.meta.update({"Dereddening with": curve_name,
                          "Dereddening with E(B-V)": Ebv})
        return spex

    @u.quantity_input(equivalencies=u.spectral())
    def rebin_spectra(self, new_waves: u.Quantity[u.AA]):
        """
        Rebin a synphot spectra to a new wavelength grid conserving flux.

        Grid does not need to be linear and can be at higher or lower
        resolution.

        Parameters
        ----------
        new_waves: an array of the output wavelenghts in Angstroms but other
        units can be specified.

        Returns
        -------
        spex : Spextrum
            New rebinned ``Spextrum`` instance.
        """
        new_waves = new_waves.to(u.AA).value

        waves = self.waveset.value  # else assumed to be angstroms
        flux = np.ones_like(waves)
        filt = SpectralElement(Empirical1D, points=waves, lookup_table=flux)
        obs = Observation(self, filt, binset=new_waves, force="taper")
        newflux = obs.binflux
        rebin_spec = Empirical1D(
            points=new_waves, lookup_table=newflux, meta=self.meta
        )
        spex = Spextrum(modelclass=rebin_spec)
        spex = self._restore_attr(spex)
        spex.meta.update({"rebinned_spectra": True})

        return spex

    def logrebin(self):
        """
        TODO: Estimate optimal rebinning factors here.
        Returns
        -------
        a Spextrum with a log rebinned in wavelength
        """
        nbins = int(self.waveset.ptp() / np.diff(self.waveset).min())
        logwaves = np.geomspace(self.wave_min, self.wave_max, nbins)
        return self.rebin_spectra(logwaves)

    def smooth(self, sigma):
        """
        Smooth the Spectrum with a Gaussian Kernel.

        Expressed in km/s (or other velocity unit).

        The spectra is first log-rebinned so it has a constant velocity spacing
        and after that is smoothed with a gaussian kernel.


        Parameters
        ----------
        sigma: u.Quantity, width of the Kernel in

        Returns
        -------
        smoothed spextrum
        """
        sigma <<= u.km / u.s

        sp_log = self.logrebin()
        step_size = np.median(np.diff(sp_log.waveset) / sp_log.waveset[:1]
                              * speed_of_light).to(u.km/u.s)

        if step_size > sigma:
            warnings.warn("Spectrum is undersampled for provided sigma value, "
                          f"{step_size} > {sigma}.")

        conv_sigma = sigma / step_size
        smoothed_flux = gaussian_filter1d(
            sp_log(sp_log.waveset), conv_sigma.value)

        self.meta.update({"KERNEL_SIZE": sigma.value})
        smoothed = Empirical1D(
            points=sp_log.waveset.value, lookup_table=smoothed_flux,
            meta=self.meta)
        spex = self._restore_attr(Spextrum(modelclass=smoothed))
        return spex

    @u.quantity_input(equivalencies=u.spectral())
    def add_noise(self, wmin: u.Quantity[u.AA], wmax: u.Quantity[u.AA]):
        """Return spectrum with given S/N in the specified wavelength range."""
        raise NotImplementedError()

    def _restore_attr(self, spextrum):
        """
        just a hack to down stream important attributes in particular methods.
        there must be a better way
        """
        temp_dict = {
            key: value
            for key, value in self.__dict__.items()
            if not key.startswith("_")
        }
        spextrum.__dict__.update(temp_dict)
        return spextrum

    # ------ Copied from synphot.SourceSpectrum so operations can also happen here --------
    # docstrings modified...

    def __add__(self, other):
        """Add `self` with `other`."""
        self._validate_other_add_sub(other)
        result = self.__class__(modelclass=self.model + other.model)
        self._merge_meta(self, other, result)
        return result

    def __sub__(self, other):
        """Subtract `other` from `self`."""
        self._validate_other_add_sub(other)
        result = self.__class__(modelclass=self.model - other.model)
        self._merge_meta(self, other, result)
        return result

    def __mul__(self, other):
        """Multiply `self` and `other`."""
        self._validate_other_mul_div(other)

        if isinstance(other, (u.Quantity, numbers.Number)):
            newcls = self.__class__(modelclass=self.model | Scale(other))
        elif isinstance(other, BaseUnitlessSpectrum):
            newcls = self.__class__(modelclass=self.model * other.model)
        else:  # Source spectrum
            raise exceptions.IncompatibleSources(
                "Cannot multiply two source spectra together."
            )

        self._merge_meta(self, other, newcls)
        return newcls

    def _validate_other_add_sub(self, other):
        """
        Check conditions for other to satisfy before add/sub.

        Copied from ``synphot.SourceSpectrum`` so additions can happen with a
        ``SourceSpectrum``.
        """
        if not isinstance(other, (self.__class__, SourceSpectrum)):
            raise exceptions.IncompatibleSources(
                f"Can only operate on {self.__class__.__name__}."
            )

    def _validate_other_mul_div(self, other):
        """
        Check conditions for other to satisfy before mul/div.

        Copied from ``synphot.SourceSpectrum`` so multiplications can happen
        with a ``SourceSpectrum``.
        """

        if not isinstance(
            other,
            (
                u.Quantity,
                numbers.Number,
                BaseUnitlessSpectrum,
                SourceSpectrum,
                self.__class__,
            ),
        ):
            raise exceptions.IncompatibleSources(
                "Can only operate on scalar number/Quantity or spectrum."
            )
        if isinstance(other, u.Quantity) and (
            other.unit.decompose() != u.dimensionless_unscaled
            or not np.isscalar(other.value)
            or not isinstance(other.value, numbers.Real)
        ):
            raise exceptions.IncompatibleSources(
                "Can only operate on real scalar dimensionless Quantity."
            )
        if isinstance(other, numbers.Number) and not (
            np.isscalar(other) and isinstance(other, numbers.Real)
        ):
            raise exceptions.IncompatibleSources(
                "Can only operate on real scalar number."
            )

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f"{self.__class__.__name__}{self.repr}"


# ------------------------------ END    -------------------------------------------


def get_vega_spectrum():
    """
    Retrieve the Vega spectrum from the database.

    Notes
    -----
    To access wavelength and fluxes use::
        spex = get_vega_spectrum()
        waves, fluxes = spex.waveset, spex(spex.waveset)

    """
    return Spextrum("ref/vega")


def _default_waves():
    """Generate default waveset with R~805."""
    waves, _ = utils.generate_wavelengths(
        minwave=100, maxwave=50000, num=5000, log=True, wave_unit=u.AA
    )
    return waves


def _read_spec(fname, fmt, **kwargs):
    if fmt == "fits":
        return read_fits_spec(fname, ext=1, **kwargs)
    elif fmt == "ascii":
        kwargs = {
            key: kwargs[key]
            for key in kwargs.keys() & {"wave_unit", "flux_unit"}
        }
        return read_ascii_spec(fname, **kwargs)
    else:
        raise SpextraError("invalid data type")


def _angstrom_value(value):
    if isinstance(value, u.Quantity):
        return value.to(u.AA, equivalencies=u.spectral()).value
    return value



# -*- coding: utf-8 -*-
"""Utility functions for downloads. Subject to change."""

import warnings
from pathlib import Path

import pooch

from .configuration import config


__all__ = ["retriever",
           "download_file",
           "download_svo_filter"]


retriever = pooch.create(path=config.cache_dir,
                         base_url=config.database_url,
                         retry_if_failed=config.retry)
retriever.load_registry(config.registry_file)


def download_file(remote_url, local_name):
    """For backwards compatibility only."""
    warnings.warn("The download_file function is deprecated and will be "
                  "removed in v1.0. Please use retriever.fetch instead.",
                  DeprecationWarning, stacklevel=2)
    file = pooch.retrieve(remote_url, known_hash=None,
                          fname=local_name.name, path=local_name.parent)
    return file


def download_svo_filter(filter_name):
    """
    Query the SVO service for the true transmittance for a given filter.

    Parameters
    ----------
    filter_name : str
        Name of the filter as available on the spanish VO filter service
        e.g: ``Paranal/HAWKI.Ks``

    Returns
    -------
    filt_curve : tuple with wave and trans values
    """
    from astropy.table import Table

    data_dir = config.cache_dir

    origin = ("http://svo2.cab.inta-csic.es/theory/fps3/"
              f"fps.php?ID={filter_name}")

    local_path_cache = Path(data_dir, "svo_filters", filter_name)
    local_path_package = config.cache_dir / "svo" / filter_name

    if local_path_package.exists():
        local_path = local_path_package
    else:
        local_path = local_path_cache
        if not local_path.exists():
            download_file(origin, local_path)

    # raises ValueError if table is malformed
    # this can be used to catch problmes
    tbl = Table.read(local_path, format="votable")

    wave = tbl["Wavelength"]
    trans = tbl["Transmission"]

    return wave, trans
